/*
 * test.c - test a portable implementation of snprintf
 *
 * AUTHOR: Mark Martinec <mark.martinec@ijs.si>, April 1999
 * Copyright: Mark Martinec
 *
 * Terms and conditions for copying,
 * distribution and modification, NO WARRANTY:
 *   GNU GENERAL PUBLIC LICENSE, Version 2 or later
 *   (see file LICENSE in the distribution kit)
 *
 * NOTE: This test program is a quick and dirty tool
 *       used while testing and benchmarking my portable snprintf.
 *       Certain data types are not fully supported, certain test
 *       cases were fabricated during testing by modifying the code
 *       or running it by specifying test parameters in the command line.
 *
 *       You are on your own if you want to use this test program!
 */

/* If no command arguments are specified do the exhaustive test.
 * This takes a long time. You may want to reduce the fw and fp
 * upper limits in the for loops.
 * You may also reduce the number of test elements in the array iargs.
 */

#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

#include <assert.h>
#include <time.h>

#ifdef HAVE_SNPRINTF
extern int portable_snprintf(char *str, size_t str_m, const char *fmt, /*args*/ ...);
#else
extern int snprintf(char *, size_t, const char *, /*args*/ ...);
#endif

#ifndef HAVE_SNPRINTF
#define portable_snprintf snprintf
#endif

#define min(a,b) ((a)<(b) ? (a) : (b))
#define max(a,b) ((a)>(b) ? (a) : (b))

int main(int argc, char *argv[]) {
  char str1[256], str2[256];
#ifdef HAVE_SNPRINTF
  char str3[256];
#endif
  int len1, len2, len3;
  int bad = 0;
  size_t str_m = 20; /* declared str size */

  if (0) {
 /* benchmarking */
    const int cnt = 100000;
    size_t size;
    char str[40000];
    time_t t0,t;
    int j,len;
    char *p;

    size = 18000;

    printf("\n\nsize = %d\n", (int)size);
    p = malloc(size); assert(p);
    memset(p,'h',size); p[size-1] = '\0';
    t0 = clock();

    printf("\nmeasuring time to SKIP a long format with no conversions\n");
    p[0] = '%'; p[1] = 's';
    for (j=cnt; j>0; j--) portable_snprintf(str,(size_t)8,p,"1234567890");
    t = clock(); printf("t_port   = %1.3f s\n", (float)(t-t0)/CLOCKS_PER_SEC); t0 = t;
    p[0] = p[1] = 'h';

    printf("\nmeasuring time to copy a long format with no conversions\n");
    for (j=cnt; j>0; j--) portable_snprintf(str,sizeof(str),p);
    t = clock(); printf("t_port   = %1.3f s\n", (float)(t-t0)/CLOCKS_PER_SEC); t0 = t;
    for (j=cnt; j>0; j--) sprintf(str,p);
    t = clock(); printf("t_sys    = %1.3f s\n", (float)(t-t0)/CLOCKS_PER_SEC); t0 = t;

    printf("\nmeasuring time to copy a long format with one conversion\n");
    p[size-10] = '%';
    for (j=cnt; j>0; j--) portable_snprintf(str,sizeof(str),p);
    t = clock(); printf("t_port   = %1.3f s\n", (float)(t-t0)/CLOCKS_PER_SEC); t0 = t;
    for (j=cnt; j>0; j--) sprintf(str,p);
    t = clock(); printf("t_sys    = %1.3f s\n", (float)(t-t0)/CLOCKS_PER_SEC); t0 = t;

    printf("\nmeasuring string argument copy speed\n");
    for (j=cnt; j>0; j--) portable_snprintf(str,sizeof(str),"%.18000s",p);
    t = clock(); printf("t_port   = %1.3f s\n", (float)(t-t0)/CLOCKS_PER_SEC); t0 = t;
    for (j=cnt; j>0; j--) sprintf(str,"%.18000s",p);
    t = clock(); printf("t_sys    = %1.3f s\n", (float)(t-t0)/CLOCKS_PER_SEC); t0 = t;

    printf("\nmeasuring left padding speed\n");
    p[0] = '\0';
    for (j=cnt; j>0; j--) portable_snprintf(str,sizeof(str),"%-18000s",p);
    t = clock(); printf("t_port   = %1.3f s\n", (float)(t-t0)/CLOCKS_PER_SEC); t0 = t;
    for (j=cnt; j>0; j--) sprintf(str,"%-18000s",p);
    t = clock(); printf("t_sys    = %1.3f s\n", (float)(t-t0)/CLOCKS_PER_SEC); t0 = t;

    printf("\nmeasuring right padding speed\n");
    p[0] = '\0';
    for (j=cnt; j>0; j--) portable_snprintf(str,sizeof(str),"%18000s",p);
    t = clock(); printf("t_port   = %1.3f s\n", (float)(t-t0)/CLOCKS_PER_SEC); t0 = t;
    for (j=cnt; j>0; j--) sprintf(str,"%18000s",p);
    t = clock(); printf("t_sys    = %1.3f s\n", (float)(t-t0)/CLOCKS_PER_SEC); t0 = t;

    printf("\nmeasuring zero padding speed\n");
    for (j=cnt; j>0; j--) portable_snprintf(str,sizeof(str),"%018000d",1);
    t = clock(); printf("t_port   = %1.3f s\n", (float)(t-t0)/CLOCKS_PER_SEC); t0 = t;
    for (j=cnt; j>0; j--) sprintf(str,"%018000d",1);
    t = clock(); printf("t_sys    = %1.3f s\n", (float)(t-t0)/CLOCKS_PER_SEC); t0 = t;

    printf("\nmeasuring system's sprintf to efficiently handle truncated strings\n");
    memset(p,'h',size); p[size-1] = '\0';
    t0 = clock();
    for (j=cnt; j>0; j--) len = strlen(p);
    printf("len = %d\n", len);
    t = clock(); printf("t_strlen = %1.3f s\n", (float)(t-t0)/CLOCKS_PER_SEC); t0 = t;
/* test if the system sprintf scans the whole string (e.g. by strlen)
 * before recognizing this wasa bad idea since the format specified
 * a truncated string precision, e.g. "%.8s" .
 */
    for (j=cnt; j>0; j--) sprintf(str,"%.2s",p);
    t = clock(); printf("t_sys    = %1.3f s\n", (float)(t-t0)/CLOCKS_PER_SEC); t0 = t;
#ifdef HAVE_SNPRINTF
    for (j=cnt; j>0; j--) snprintf(str,sizeof(str),"%.2s",p);
    t = clock(); printf("t_sys    = %1.3f s\n", (float)(t-t0)/CLOCKS_PER_SEC); t0 = t;
#endif
    for (j=cnt; j>0; j--) portable_snprintf(str,sizeof(str),"%.2s",p);
    t = clock(); printf("t_port   = %1.3f s\n", (float)(t-t0)/CLOCKS_PER_SEC); t0 = t;

    free(p);
    return 0;
  }

 /* testing for correctness and compatibility */
  if (argc >= 3) {
    char *c; int alldigits = 1;
    for (c=argv[2]; *c; c++)
      if (! (*c == '-' || (*c >= '0' && *c <= '9'))) alldigits = 0;
    if (alldigits) {
      int j = atoi(argv[2]);
      len1 = portable_snprintf(str1, str_m, argv[1], j, 3);
      len2 = sprintf(str2,                  argv[1], j, 3);
#ifdef HAVE_SNPRINTF
      len3 = snprintf(str3, str_m,          argv[1], j, 3);
#endif
    } else {
      len1 = portable_snprintf(str1, str_m, argv[1], argv[2], 3);
      len2 = sprintf(str2,                  argv[1], argv[2], 3);
#ifdef HAVE_SNPRINTF
      len3 = snprintf(str3, str_m,          argv[1], argv[2], 3);
#endif
    }
    printf("portable:     |%s|  len = %d\n", str1, len1);
    printf("sys sprintf:  |%s|  len = %d\n", str2, len2);
#ifdef HAVE_SNPRINTF
    printf("sys snprintf: |%s|  len = %d\n", str3, len3);
#endif
  } else {  /* exhaustive testing */

    const char flags[] = "+- 0#";     /* set of test flags   (including '\0')*/
    int flags_l = strlen(flags);

    const char fspec[] = "scdpoxXuiy"; /* set of test formats (including '\0') */
    int fspec_l = strlen(fspec);

#ifdef SNPRINTF_LONGLONG_SUPPORT
    const char datatype[] = " hl2"; /* set of datatypes */
#else
    const char datatype[] = " hl";  /* set of datatypes */
#endif
    int datatype_l = strlen(datatype);

    const long int iargs[] =                /* set of numeric test arguments */
      { 0,1,9,10,28,99,100,127,128,129,998,1000,32767,32768,32769,
        -1,-9,-10,-28,-99,-100,-127,-128,-129,
        -998,-1000,-32767,-32768,-32769 };
    int iargs_l = sizeof(iargs)/sizeof(iargs[0]);
    const char *sargs[] =             /* set of string test arguments */
      { "", "a", "0", "-ab", "abcde", "abcdefghijklmnopqrstuv" };
    int sargs_l = sizeof(sargs)/sizeof(sargs[0]);

    char fmt[256];
    int fmt_l;
    int a, fs, fl1, fl2, fl3, fl4, fl5, fw, fp, dt;

    for (fs=0; fs<=fspec_l; fs++) {  /* format specifier */
      int strtype = (fspec[fs] == 's' || fspec[fs] == '%');
      int args_l = (strtype ? sargs_l : iargs_l);

      for (fw=-1; fw<=6; fw++) {  /* minimal field width */

        printf("Trying format  %%");
        if (fw >= 0) printf("%d", fw);
        if (fspec[fs]) putchar(fspec[fs]);
        putchar('\n');

        for (fp=-2; fp<=6; fp++) {   /* format field precision */

          /* data type modifiers */
          for (dt=0; dt < ((strtype||fspec[fs]=='c') ? 1 : datatype_l); dt++) {

            int dataty = datatype[dt];

            if (fspec[fs] == 'D' || fspec[fs] == 'U' || fspec[fs] == 'O')
              dataty = 'l';

            if (fspec[fs] == 'p' && dataty == '2') continue;

            for (fl1=0; fl1<=flags_l; fl1++) {  /* flags */
             for (fl2=0; fl2<=flags_l; fl2++) {
              for (fl3=0; fl3<=flags_l; fl3++) {
               for (fl4=0; fl4<=flags_l; fl4++) {
                for (fl5=0; fl5<=flags_l; fl5++) {

                   for (a=0; a<args_l; a++) {  /* test arguments */

                     fmt_l = 0; fmt[fmt_l++] = '%';
                     if (flags[fl1]) fmt[fmt_l++] = flags[fl1];
                     if (flags[fl2]) fmt[fmt_l++] = flags[fl2];
                     if (flags[fl3]) fmt[fmt_l++] = flags[fl3];
                     if (flags[fl4]) fmt[fmt_l++] = flags[fl4];
                     if (flags[fl5]) fmt[fmt_l++] = flags[fl5];
                     if (fw >= 0) fmt_l += sprintf(fmt+fmt_l, "%d", fw);
                     if (fp >= -1) {
                       fmt[fmt_l++] = '.';
                       if (fp >= 0) fmt_l += sprintf(fmt+fmt_l, "%d", fp);
                     }
                     if (dataty == '2')
                       { fmt[fmt_l++] = 'l'; fmt[fmt_l++] = 'l'; }
                     else if (dataty != ' ')
                       { fmt[fmt_l++] = dataty; }

                     if (fspec[fs]) fmt[fmt_l++] = fspec[fs];
                     fmt[fmt_l++] = '\0';

                     if (a==0 && fl1==flags_l && fl2==flags_l && fl3==flags_l
                         && fl4==flags_l && fl5==flags_l) printf("%s\n", fmt);

#ifdef HAVE_SNPRINTF
                     memset(str1,'G',sizeof(str1));
                     memset(str2,'G',sizeof(str2));
                     memset(str3,'G',sizeof(str3));
#endif
                     len1 = len2 = len3 = 0;
                     if (strtype) {
                       len1 = portable_snprintf(str1, str_m, fmt, sargs[a]);
                       len2 = sprintf(str2,                  fmt, sargs[a]);
#ifdef HAVE_SNPRINTF
                       len3 = snprintf(str3, str_m,          fmt, sargs[a]);
#endif
                     } else if (fspec[fs] == 'p') {
                       len1 = portable_snprintf(str1, str_m, fmt, (void *)iargs[a]);
                       len2 = sprintf(str2,                  fmt, (void *)iargs[a]);
#ifdef HAVE_SNPRINTF
                       len3 = snprintf(str3, str_m,          fmt, (void *)iargs[a]);
#endif
                     } else {
                       switch (dataty) {
                       case '\0':
                         len1 = portable_snprintf(str1, str_m, fmt, (int) iargs[a]);
                         len2 = sprintf(str2,                  fmt, (int) iargs[a]);
#ifdef HAVE_SNPRINTF
                         len3 = snprintf(str3, str_m,          fmt, (int) iargs[a]);
#endif
                         break;
                       case 'h':
                         len1 = portable_snprintf(str1, str_m, fmt, (short int)iargs[a]);
                         len2 = sprintf(str2,                  fmt, (short int)iargs[a]);
#ifdef HAVE_SNPRINTF
                         len3 = snprintf(str3, str_m,          fmt, (short int)iargs[a]);
#endif
                         break;
                       case 'l':
                         len1 = portable_snprintf(str1, str_m, fmt, (long int)iargs[a]);
                         len2 = sprintf(str2,                  fmt, (long int)iargs[a]);
#ifdef HAVE_SNPRINTF
                         len3 = snprintf(str3, str_m,          fmt, (long int)iargs[a]);
#endif
                         break;
#ifdef SNPRINTF_LONGLONG_SUPPORT
                       case '2':
                         len1 = portable_snprintf(str1, str_m, fmt, (long long int)iargs[a]);
                         len2 = sprintf(str2,                  fmt, (long long int)iargs[a]);
#ifdef HAVE_SNPRINTF
                         len3 = snprintf(str3, str_m,          fmt, (long long int)iargs[a]);
#endif
                         break;
#endif
                       }
                     }
                     if (0) {
#ifdef HAVE_SNPRINTF
                     } else if (len1 != len3 ||
                         memcmp(str1,str3,min(len1+20,sizeof(str1))) != 0) {
                       bad = 1;
                       if (strtype) printf("\n2: %s, <%s>\n", fmt, sargs[a]);
                       else         printf("\n2: %s, %ld\n",   fmt, iargs[a]);
                       printf("portable:     |%s|  len = %d\n", str1, len1);
                       printf("sys sprintf:  |%s|  len = %d\n", str2, len2);
                       printf("sys snprintf: |%s|  len = %d\n", str3, len3);
#else
                     } else if (len1 != len2 ||
                         (len1>0 && memcmp(str1,str2,min(len1,str_m)-1) != 0)) {
                       bad = 1;
                       if (strtype) printf("\n1: %s, <%s>\n", fmt, sargs[a]);
                       else         printf("\n1: %s, %ld\n",   fmt, iargs[a]);
                       printf("portable:     |%s|  len = %d\n", str1, len1);
                       printf("sys sprintf:  |%s|  len = %d\n", str2, len2);
#endif
                     }
                     if (bad) return(1);
                   }

                }
               }
              }
             }

            }
          }
        }
      }
    }
  }
  return (bad?1:0);
}
